@@ -20,17 +20,18 @@ use std::{
     },
 };
 
-use async_lock::RwLock;
+use arc_bytes::OwnedBytes;
 pub use flume;
+use parking_lot::RwLock;
 use serde::{Deserialize, Serialize};
 
 /// A `PubSub` message.
 #[derive(Clone, Debug, Serialize, Deserialize)]
 pub struct Message {
     /// The topic of the message.
-    pub topic: String,
+    pub topic: OwnedBytes,
     /// The payload of the message.
-    pub payload: Vec<u8>,
+    pub payload: OwnedBytes,
 }
 
 impl Message {
@@ -39,19 +40,36 @@ impl Message {
     /// # Errors
     ///
     /// Returns an error if `payload` fails to serialize with `pot`.
-    pub fn new<S: Into<String>, P: Serialize>(topic: S, payload: &P) -> Result<Self, pot::Error> {
-        Ok(Self {
-            topic: topic.into(),
-            payload: pot::to_vec(payload)?,
-        })
+    pub fn new<Topic: Serialize, Payload: Serialize>(
+        topic: &Topic,
+        payload: &Payload,
+    ) -> Result<Self, pot::Error> {
+        Ok(Self::raw(pot::to_vec(topic)?, pot::to_vec(payload)?))
     }
 
-    /// Deserialize the payload as `P` using `pot`.
+    /// Creates a new message with a raw payload.
+    pub fn raw<S: Into<OwnedBytes>, B: Into<OwnedBytes>>(topic: S, payload: B) -> Self {
+        Self {
+            topic: topic.into(),
+            payload: payload.into(),
+        }
+    }
+
+    /// Deserialize the topic as `Topic` using `pot`.
+    ///
+    /// # Errors
+    ///
+    /// Returns an error if `topic` fails to deserialize with `pot`.
+    pub fn topic<'a, Topic: Deserialize<'a>>(&'a self) -> Result<Topic, pot::Error> {
+        pot::from_slice(&self.topic).map_err(pot::Error::from)
+    }
+
+    /// Deserialize the payload as `Payload` using `pot`.
     ///
     /// # Errors
     ///
     /// Returns an error if `payload` fails to deserialize with `pot`.
-    pub fn payload<P: for<'de> Deserialize<'de>>(&self) -> Result<P, pot::Error> {
+    pub fn payload<'a, Payload: Deserialize<'a>>(&'a self) -> Result<Payload, pot::Error> {
         pot::from_slice(&self.payload).map_err(pot::Error::from)
     }
 }
@@ -68,7 +86,7 @@ pub struct Relay {
 #[derive(Debug, Default)]
 struct Data {
     subscribers: RwLock<HashMap<SubscriberId, SubscriberInfo>>,
-    topics: RwLock<HashMap<String, TopicId>>,
+    topics: RwLock<HashMap<OwnedBytes, TopicId>>,
     subscriptions: RwLock<HashMap<TopicId, HashSet<SubscriberId>>>,
     last_topic_id: AtomicU64,
     last_subscriber_id: AtomicU64,
@@ -76,8 +94,8 @@ struct Data {
 
 impl Relay {
     /// Create a new [`Subscriber`] for this relay.
-    pub async fn create_subscriber(&self) -> Subscriber {
-        let mut subscribers = self.data.subscribers.write().await;
+    pub fn create_subscriber(&self) -> Subscriber {
+        let mut subscribers = self.data.subscribers.write();
         let id = self.data.last_subscriber_id.fetch_add(1, Ordering::SeqCst);
         let (sender, receiver) = flume::unbounded();
         subscribers.insert(
@@ -92,8 +110,6 @@ impl Relay {
                 id,
                 receiver,
                 relay: self.clone(),
-                #[cfg(not(target_arch = "wasm32"))]
-                tokio: tokio::runtime::Handle::current(),
             }),
         }
     }
@@ -102,58 +118,76 @@ impl Relay {
     ///
     /// # Errors
     ///
-    /// Returns an error if `payload` fails to serialize with `pot`.
-    pub async fn publish<S: Into<String> + Send, P: Serialize + Sync>(
+    /// Returns an error if `topic` or `payload` fails to serialize with `pot`.
+    pub fn publish<Topic: Serialize, P: Serialize>(
         &self,
-        topic: S,
+        topic: &Topic,
         payload: &P,
     ) -> Result<(), pot::Error> {
         let message = Message::new(topic, payload)?;
-        self.publish_message(message).await;
+        self.publish_message(&message);
         Ok(())
     }
 
+    /// Publishes a `payload` to all subscribers of `topic`.
+    pub fn publish_raw<Topic: Into<OwnedBytes>, Payload: Into<OwnedBytes>>(
+        &self,
+        topic: Topic,
+        payload: Payload,
+    ) {
+        let message = Message::raw(topic, payload);
+        self.publish_message(&message);
+    }
+
     /// Publishes a `payload` to all subscribers of `topic`.
     ///
     /// # Errors
     ///
-    /// Returns an error if `payload` fails to serialize with `pot`.
-    pub async fn publish_to_all<P: Serialize + Sync>(
+    /// Returns an error if `topics` or `payload` fail to serialize with `pot`.
+    pub fn publish_to_all<
+        'topics,
+        Topics: IntoIterator<Item = &'topics Topic> + 'topics,
+        Topic: Serialize + 'topics,
+        Payload: Serialize,
+    >(
         &self,
-        topics: Vec<String>,
-        payload: &P,
+        topics: Topics,
+        payload: &Payload,
     ) -> Result<(), pot::Error> {
-        let tasks = topics
-            .into_iter()
-            .map(|topic| Message::new(topic, payload).map(|message| self.publish_message(message)))
-            .collect::<Result<Vec<_>, _>>()?;
-        futures::future::join_all(tasks).await;
+        for topic in topics {
+            let message = Message::new(topic, payload)?;
+            self.publish_message(&message);
+        }
 
         Ok(())
     }
 
     /// Publishes a `payload` to all subscribers of `topic`.
-    pub async fn publish_serialized_to_all(&self, topics: Vec<String>, payload: Vec<u8>) {
-        let tasks = topics.into_iter().map(|topic| {
-            self.publish_message(Message {
+    pub fn publish_raw_to_all(
+        &self,
+        topics: impl IntoIterator<Item = OwnedBytes>,
+        payload: impl Into<OwnedBytes>,
+    ) {
+        let payload = payload.into();
+        for topic in topics {
+            self.publish_message(&Message {
                 topic,
                 payload: payload.clone(),
-            })
-        });
-        futures::future::join_all(tasks).await;
-    }
-
-    /// Publishes a message to all subscribers of its topic.
-    pub async fn publish_message(&self, message: Message) {
-        if let Some(topic_id) = self.topic_id(&message.topic).await {
-            self.post_message_to_topic(message, topic_id).await;
+            });
         }
     }
 
-    async fn add_subscriber_to_topic(&self, subscriber_id: u64, topic: String) {
-        let mut subscribers = self.data.subscribers.write().await;
-        let mut topics = self.data.topics.write().await;
-        let mut subscriptions = self.data.subscriptions.write().await;
+    /// Publishes a message to all subscribers of its topic.
+    pub fn publish_message(&self, message: &Message) {
+        if let Some(topic_id) = self.topic_id(&message.topic) {
+            self.post_message_to_topic(message, topic_id);
+        }
+    }
+
+    fn add_subscriber_to_topic(&self, subscriber_id: u64, topic: OwnedBytes) {
+        let mut subscribers = self.data.subscribers.write();
+        let mut topics = self.data.topics.write();
+        let mut subscriptions = self.data.subscriptions.write();
 
         // Lookup or create a topic id
         let topic_id = *topics
@@ -168,9 +202,9 @@ impl Relay {
         subscribers.insert(subscriber_id);
     }
 
-    async fn remove_subscriber_from_topic(&self, subscriber_id: u64, topic: &str) {
-        let mut subscribers = self.data.subscribers.write().await;
-        let mut topics = self.data.topics.write().await;
+    fn remove_subscriber_from_topic(&self, subscriber_id: u64, topic: &[u8]) {
+        let mut subscribers = self.data.subscribers.write();
+        let mut topics = self.data.topics.write();
 
         let remove_topic = if let Some(topic_id) = topics.get(topic) {
             if let Some(subscriber) = subscribers.get_mut(&subscriber_id) {
@@ -183,7 +217,7 @@ impl Relay {
                 return;
             }
 
-            let mut subscriptions = self.data.subscriptions.write().await;
+            let mut subscriptions = self.data.subscriptions.write();
             let remove_topic = if let Some(subscriptions) = subscriptions.get_mut(topic_id) {
                 subscriptions.remove(&subscriber_id);
                 subscriptions.is_empty()
@@ -206,30 +240,32 @@ impl Relay {
         }
     }
 
-    async fn topic_id(&self, topic: &str) -> Option<TopicId> {
-        let topics = self.data.topics.read().await;
+    fn topic_id(&self, topic: &[u8]) -> Option<TopicId> {
+        let topics = self.data.topics.read();
         topics.get(topic).copied()
     }
 
-    async fn post_message_to_topic(&self, message: Message, topic: TopicId) {
+    fn post_message_to_topic(&self, message: &Message, topic: TopicId) {
         let failures = {
             // For an optimal-flow case, we're going to acquire read permissions
             // only, allowing messages to be posted in parallel. This block is
             // responsible for returning `SubscriberId`s that failed to send.
-            let subscribers = self.data.subscribers.read().await;
-            let subscriptions = self.data.subscriptions.read().await;
+            let subscribers = self.data.subscribers.read();
+            let subscriptions = self.data.subscriptions.read();
             if let Some(registered) = subscriptions.get(&topic) {
-                let message = Arc::new(message);
-                let failures = futures::future::join_all(registered.iter().filter_map(|id| {
-                    subscribers.get(id).map(|subscriber| {
-                        let message = message.clone();
-                        async move { (*id, subscriber.sender.send_async(message).await.is_ok()) }
+                let failures = registered
+                    .iter()
+                    .filter_map(|id| {
+                        subscribers.get(id).and_then(|subscriber| {
+                            let message = message.clone();
+                            if subscriber.sender.send(message).is_ok() {
+                                None
+                            } else {
+                                Some(*id)
+                            }
+                        })
                     })
-                }))
-                .await
-                .into_iter()
-                .filter_map(|(id, sent_message)| if sent_message { None } else { Some(id) })
-                .collect::<Vec<SubscriberId>>();
+                    .collect::<Vec<SubscriberId>>();
 
                 failures
             } else {
@@ -239,15 +275,15 @@ impl Relay {
 
         if !failures.is_empty() {
             for failed in failures {
-                self.unsubscribe_all(failed).await;
+                self.unsubscribe_all(failed);
             }
         }
     }
 
-    async fn unsubscribe_all(&self, subscriber_id: SubscriberId) {
-        let mut subscribers = self.data.subscribers.write().await;
-        let mut topics = self.data.topics.write().await;
-        let mut subscriptions = self.data.subscriptions.write().await;
+    fn unsubscribe_all(&self, subscriber_id: SubscriberId) {
+        let mut subscribers = self.data.subscribers.write();
+        let mut topics = self.data.topics.write();
+        let mut subscriptions = self.data.subscriptions.write();
         if let Some(subscriber) = subscribers.remove(&subscriber_id) {
             for topic in &subscriber.topics {
                 let remove = if let Some(subscriptions) = subscriptions.get_mut(topic) {
@@ -268,36 +304,57 @@ impl Relay {
 
 #[derive(Debug)]
 struct SubscriberInfo {
-    sender: flume::Sender<Arc<Message>>,
+    sender: flume::Sender<Message>,
     topics: HashSet<u64>,
 }
 
 /// A subscriber for [`Message`]s published to subscribed topics.
 #[derive(Debug, Clone)]
+#[must_use]
 pub struct Subscriber {
     data: Arc<SubscriberData>,
 }
 
 impl Subscriber {
     /// Subscribe to [`Message`]s published to `topic`.
-    pub async fn subscribe_to<S: Into<String> + Send>(&self, topic: S) {
+    ///
+    /// # Errors
+    ///
+    /// Returns an error if `topic` fails to serialize with `pot`.
+    pub fn subscribe_to<Topic: Serialize>(&self, topic: &Topic) -> Result<(), pot::Error> {
+        let topic = pot::to_vec(topic)?;
+        self.subscribe_to_raw(topic);
+        Ok(())
+    }
+
+    /// Subscribe to [`Message`]s published to `topic`.
+    pub fn subscribe_to_raw(&self, topic: impl Into<OwnedBytes>) {
         self.data
             .relay
-            .add_subscriber_to_topic(self.data.id, topic.into())
-            .await;
+            .add_subscriber_to_topic(self.data.id, topic.into());
     }
 
     /// Unsubscribe from [`Message`]s published to `topic`.
-    pub async fn unsubscribe_from(&self, topic: &str) {
+    ///
+    /// # Errors
+    ///
+    /// Returns an error if `topic` fails to serialize with `pot`.
+    pub fn unsubscribe_from<Topic: Serialize>(&self, topic: &Topic) -> Result<(), pot::Error> {
+        let topic = pot::to_vec(topic)?;
+        self.unsubscribe_from_raw(&topic);
+        Ok(())
+    }
+
+    /// Unsubscribe from [`Message`]s published to `topic`.
+    pub fn unsubscribe_from_raw(&self, topic: &[u8]) {
         self.data
             .relay
-            .remove_subscriber_from_topic(self.data.id, topic)
-            .await;
+            .remove_subscriber_from_topic(self.data.id, topic);
     }
 
     /// Returns the receiver to receive [`Message`]s.
     #[must_use]
-    pub fn receiver(&self) -> &'_ flume::Receiver<Arc<Message>> {
+    pub fn receiver(&self) -> &'_ flume::Receiver<Message> {
         &self.data.receiver
     }
 
@@ -312,22 +369,12 @@ impl Subscriber {
 struct SubscriberData {
     id: SubscriberId,
     relay: Relay,
-    receiver: flume::Receiver<Arc<Message>>,
-    #[cfg(not(target_arch = "wasm32"))]
-    tokio: tokio::runtime::Handle,
+    receiver: flume::Receiver<Message>,
 }
 
 impl Drop for SubscriberData {
     fn drop(&mut self) {
-        let id = self.id;
-        let relay = self.relay.clone();
-        let drop_task = async move {
-            relay.unsubscribe_all(id).await;
-        };
-        #[cfg(target_arch = "wasm32")]
-        wasm_bindgen_futures::spawn_local(drop_task);
-        #[cfg(not(target_arch = "wasm32"))]
-        self.tokio.spawn(drop_task);
+        self.relay.unsubscribe_all(self.id);
     }
 }
 
@@ -338,12 +385,12 @@ mod tests {
     #[tokio::test]
     async fn simple_pubsub_test() -> anyhow::Result<()> {
         let pubsub = Relay::default();
-        let subscriber = pubsub.create_subscriber().await;
-        subscriber.subscribe_to("mytopic").await;
-        pubsub.publish("mytopic", &String::from("test")).await?;
+        let subscriber = pubsub.create_subscriber();
+        subscriber.subscribe_to(&"mytopic")?;
+        pubsub.publish(&"mytopic", &String::from("test"))?;
         let receiver = subscriber.receiver().clone();
         let message = receiver.recv_async().await.expect("No message received");
-        assert_eq!(message.topic, "mytopic");
+        assert_eq!(message.topic::<String>()?, "mytopic");
         assert_eq!(message.payload::<String>()?, "test");
         // The message should only be received once.
         assert!(matches!(
@@ -359,27 +406,27 @@ mod tests {
     #[tokio::test]
     async fn multiple_subscribers_test() -> anyhow::Result<()> {
         let pubsub = Relay::default();
-        let subscriber_a = pubsub.create_subscriber().await;
-        let subscriber_ab = pubsub.create_subscriber().await;
-        subscriber_a.subscribe_to("a").await;
-        subscriber_ab.subscribe_to("a").await;
-        subscriber_ab.subscribe_to("b").await;
+        let subscriber_a = pubsub.create_subscriber();
+        let subscriber_ab = pubsub.create_subscriber();
+        subscriber_a.subscribe_to(&"a")?;
+        subscriber_ab.subscribe_to(&"a")?;
+        subscriber_ab.subscribe_to(&"b")?;
 
-        pubsub.publish("a", &String::from("a1")).await?;
-        pubsub.publish("b", &String::from("b1")).await?;
-        pubsub.publish("a", &String::from("a2")).await?;
+        pubsub.publish(&"a", &String::from("a1"))?;
+        pubsub.publish(&"b", &String::from("b1"))?;
+        pubsub.publish(&"a", &String::from("a2"))?;
 
         // Check subscriber_a for a1 and a2.
-        let message = subscriber_a.receiver().recv_async().await?;
+        let message = subscriber_a.receiver().recv()?;
         assert_eq!(message.payload::<String>()?, "a1");
-        let message = subscriber_a.receiver().recv_async().await?;
+        let message = subscriber_a.receiver().recv()?;
         assert_eq!(message.payload::<String>()?, "a2");
 
-        let message = subscriber_ab.receiver().recv_async().await?;
+        let message = subscriber_ab.receiver().recv()?;
         assert_eq!(message.payload::<String>()?, "a1");
-        let message = subscriber_ab.receiver().recv_async().await?;
+        let message = subscriber_ab.receiver().recv()?;
         assert_eq!(message.payload::<String>()?, "b1");
-        let message = subscriber_ab.receiver().recv_async().await?;
+        let message = subscriber_ab.receiver().recv()?;
         assert_eq!(message.payload::<String>()?, "a2");
 
         Ok(())
@@ -388,19 +435,19 @@ mod tests {
     #[tokio::test]
     async fn unsubscribe_test() -> anyhow::Result<()> {
         let pubsub = Relay::default();
-        let subscriber = pubsub.create_subscriber().await;
-        subscriber.subscribe_to("a").await;
+        let subscriber = pubsub.create_subscriber();
+        subscriber.subscribe_to(&"a")?;
 
-        pubsub.publish("a", &String::from("a1")).await?;
-        subscriber.unsubscribe_from("a").await;
-        pubsub.publish("a", &String::from("a2")).await?;
-        subscriber.subscribe_to("a").await;
-        pubsub.publish("a", &String::from("a3")).await?;
+        pubsub.publish(&"a", &String::from("a1"))?;
+        subscriber.unsubscribe_from(&"a")?;
+        pubsub.publish(&"a", &String::from("a2"))?;
+        subscriber.subscribe_to(&"a")?;
+        pubsub.publish(&"a", &String::from("a3"))?;
 
         // Check subscriber_a for a1 and a2.
-        let message = subscriber.receiver().recv_async().await?;
+        let message = subscriber.receiver().recv()?;
         assert_eq!(message.payload::<String>()?, "a1");
-        let message = subscriber.receiver().recv_async().await?;
+        let message = subscriber.receiver().recv()?;
         assert_eq!(message.payload::<String>()?, "a3");
 
         Ok(())
@@ -409,26 +456,26 @@ mod tests {
     #[tokio::test]
     async fn drop_and_send_test() -> anyhow::Result<()> {
         let pubsub = Relay::default();
-        let subscriber_a = pubsub.create_subscriber().await;
-        let subscriber_to_drop = pubsub.create_subscriber().await;
-        subscriber_a.subscribe_to("a").await;
-        subscriber_to_drop.subscribe_to("a").await;
+        let subscriber_a = pubsub.create_subscriber();
+        let subscriber_to_drop = pubsub.create_subscriber();
+        subscriber_a.subscribe_to(&"a")?;
+        subscriber_to_drop.subscribe_to(&"a")?;
 
-        pubsub.publish("a", &String::from("a1")).await?;
+        pubsub.publish(&"a", &String::from("a1"))?;
         drop(subscriber_to_drop);
-        pubsub.publish("a", &String::from("a2")).await?;
+        pubsub.publish(&"a", &String::from("a2"))?;
 
         // Check subscriber_a for a1 and a2.
-        let message = subscriber_a.receiver().recv_async().await?;
+        let message = subscriber_a.receiver().recv()?;
         assert_eq!(message.payload::<String>()?, "a1");
-        let message = subscriber_a.receiver().recv_async().await?;
+        let message = subscriber_a.receiver().recv()?;
         assert_eq!(message.payload::<String>()?, "a2");
 
-        let subscribers = pubsub.data.subscribers.read().await;
+        let subscribers = pubsub.data.subscribers.read();
         assert_eq!(subscribers.len(), 1);
-        let topics = pubsub.data.topics.read().await;
-        let topic_id = topics.get("a").expect("topic not found");
-        let subscriptions = pubsub.data.subscriptions.read().await;
+        let topics = pubsub.data.topics.read();
+        let topic_id = topics.values().next().expect("topic not found");
+        let subscriptions = pubsub.data.subscriptions.read();
         assert_eq!(
             subscriptions
                 .get(topic_id)
@@ -443,17 +490,15 @@ mod tests {
     #[tokio::test]
     async fn drop_cleanup_test() -> anyhow::Result<()> {
         let pubsub = Relay::default();
-        let subscriber = pubsub.create_subscriber().await;
-        subscriber.subscribe_to("a").await;
+        let subscriber = pubsub.create_subscriber();
+        subscriber.subscribe_to(&"a")?;
         drop(subscriber);
-        // Drop spawns a task that cleans up. Give it a moment to execute.
-        tokio::time::sleep(std::time::Duration::from_millis(10)).await;
 
-        let subscribers = pubsub.data.subscribers.read().await;
+        let subscribers = pubsub.data.subscribers.read();
         assert_eq!(subscribers.len(), 0);
-        let subscriptions = pubsub.data.subscriptions.read().await;
+        let subscriptions = pubsub.data.subscriptions.read();
         assert_eq!(subscriptions.len(), 0);
-        let topics = pubsub.data.topics.read().await;
+        let topics = pubsub.data.topics.read();
         assert_eq!(topics.len(), 0);
 
         Ok(())
